// AVX (256bit register ymm0-ymm16 = 4*f32) + FMA (fused-multiply-add)
//   some CPUs? gcc8? does not support mix AVX + FMA automatically?
// LRZ SuperMUC-NG 2019-05-23
// gcc485 no avx512f support
// gcc630          -mfma -mavx512f -S = vfmadd132ps %zmm1, %zmm2, %zmm0
// gcc730          -mfma -mavx512f -S = vfmadd132ps %zmm1, %zmm2, %zmm0
// gcc820          -mfma -mavx512f -S = vfmadd132ps %zmm1, %zmm2, %zmm0
// gcc830          -mfma -mavx512f -S = vfmadd132ps %zmm1, %zmm2, %zmm0
// icc1900(gcc480) -mfma -mavx512f -S = vfmadd213ps %zmm2, %zmm1, %zmm0
// icc1900(gcc830) -mfma -mavx512f -S = vfmadd213ps %zmm2, %zmm1, %zmm0
// icc1900(gcc830) -O0 ...  vmulps + vaddps  %zmm1, %zmm0, %zmm0
// icc1900(gcc830) -O1 ...  vfmadd213ps      %zmm1, %zmm0, %zmm0
// gcc   -O2 -mavx512f -mfma         # since 4.9 -S  PC16.gcc821=?
// gcc   -O2 -march=skylake ../utils/fma_avx.c # -S  PC16.gcc821=?
// clang -O1 -mavx512f -mfma -ffp-contract=fast
// icc18  -fast -xHost # Host=Host,AVX-512, AVX2, AVX, SSE4.1
// icc19  -fast -xHost # Host=Host,SKYLAKE-AVX512,SKYLAKE
// icc                 -simd -vec -vec-report[0-5]
// icc    -par_threshold0 : 0=always vectorize , 100 only if sure (dflt)
// gcc -Ofast -march=skylake  
//__m512		512-bit vector containing 8 float   (zmm0-31)
//__m128		128-bit vector containing 8 float   (xmm0-15)
//__m256		256-bit vector containing 8 float   (ymm0-15)
//__m256d		256-bit vector containing 4 doubles
//__m256i		256-bit vector containing integers
//__m512		256-bit vector containing 8 float   (ymm0-15)
//__m512d		256-bit vector containing 4 doubles
//__m512i		256-bit vector containing integers
// https://gcc.gnu.org/onlinedocs/gcc/Vector-Extensions.html
// typedef int v4si __attribute__ ((vector_size (16))); // 4*int32=16B
#include <immintrin.h>
//typedef float __m256 __attribute__ ((__vector_size__ (32),
//                                     __may_alias__));  // 8*flt32=32B                                 
#include <stdlib.h>  // atol()
#include <stdio.h>   // printf

#pragma GCC optimize ("O2","tree-vectorize","fast-math")

__m512 float_mul_addv(__m512 a, __m512 b, __m512 c) {
    return _mm512_add_ps(_mm512_mul_ps(a, b), c);
}

__m512d double_mul_addv(__m512d a, __m512d b, __m512d c) {
    return _mm512_add_pd(_mm512_mul_pd(a, b), c);
}

// do not optimize function-call away
#pragma GCC optimize ("O0","tree-vectorize","fast-math")
int main(int argn, char *argv[]) {
  __m512 a, b, c;
  float* fa = (float*)&a;
  float* fb = (float*)&b;
  float* fc = (float*)&c;
  long x1 = ((argn>1) ? atol(argv[1]) : 0); int i; // random input
  for (i=0;i<sizeof(__m512)/sizeof(float);i++) {
     fa[i]=0.2+0.1*((x1>>i)&1);
     fb[i]=0.2+0.1*((x1>>i)&3);
     fc[i]=0.2+0.1*((x1>>i)&5);
  }
    
  __m512 r = float_mul_addv(a, b, c);
  
  float* fr = (float*)&r;
  for (i=0;i<sizeof(__m512)/sizeof(float);i++)
     printf(" %f",fr[i]);
  printf("\n");

  return 0;
}
